// Description: Module Random
//      Author: Mikko Koivisto
//     Updated: May 28, 2003

#ifndef RANDOM_H
#define RANDOM_H

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <time.h>

#define PI		3.141592654
#define LN(u)	log(u)

static bool is_save_normal;
static double save_normal;

class Random {
public:
	Random(){;}
	~Random(){;}
	static void init(){
		fprintf(stderr, "random > init \n");
		time_t t; long seed=time(&t); srand(seed);
		is_save_normal = false;
	}
	static void rperm(int *a, int n){
		for (int i = n-1; i >= 0; i --){
			int j = unif_int_rnd(0, i);
			int t = a[j]; a[j] = a[i]; a[i] = t;
		}
	}
	
	static int vec_rnd(double *w, int n){
		double s = 0;
		for (int i = 0; i < n; i ++) s += w[i];
		double u = unif_rnd(0,s);
		int j = 0; s = 0;
		while (s < u) s += w[j++];
		return j-1;
	}
	
	// ---------------------
	// Uniform distribution.
	// ---------------------
	static double unif_rnd(double low, double high){
		return low+(high-low)*((double)(rand())/(RAND_MAX + 1.0));	
	}
	static double unif_pdf(double low, double high){
		return 1.0/(high-low);	
	}
	static int unif_int_rnd(int low, int high){
		return low+(int)((high - low + 1)*((double)(rand())/(RAND_MAX + 1.0)));	
	}
	static double unif_int_pdf(int low, int high){
		return 1.0/(high - low + 1);	
	}
	// ------------------
	// Beta distribution.
	// ------------------
	static double beta(double a, double b){
		return gamma(a+b)/(gamma(a)*gamma(b));
	}
	static double beta_pdf(double a, double b, double x){
		if (x <= 0 || x >= 1) return 0; 
		double y = pow(x, a-1)*pow(1-x,b-1)/beta(a, b);
		return y;
	}
	static double beta_pdf_unnorm(double a, double b, double x){
		if (x <= 0 || x >= 1) return 0; 
		double y = pow(x, a-1)*pow(1-x,b-1);
		return y;
	}
	static double beta_rnd(double a, double b){
		double a1 = gamm_rnd(a,1);
		double a1d = gamm_rnd(b,1);
		double rnd = a1/(a1+a1d);
		return rnd;
	}
	// -------------------
	// Gamma distribution.
	// -------------------
	static double gamma(double k){
		double t = 1;
		while (k > 0){t *= k; k --;}
		return t;
	}
	static double gamm_pdf(double m, double k, double x){
		double y = m*pow(m*x, k-1)*exp(-m*x)/gamma(k);
		return y;
	}
	static double gamm_rnd(double m, double k){
		double x = -1;
		if (m <= 1){
  			// [Matlab:] Use RGS algorithm by Best, p. 426
  			double c = 1/m; 
  			double t = 0.07+0.75*sqrt(1-m);
  			double b = 1+exp(-t)*m/t;
       		int accept=0;
			while (accept==0){
        		double u = unif_rnd(0,1); 
				double w = unif_rnd(0,1); 
				double v = b*u;
          		if (v <= 1){
            		x = t*(pow(v,c));
             		accept = ((w<=((2-x)/(2+x))) || (w<=exp(-x)));
				}
				else{
					x = -log(c*t*(b-v));
             		double y = x/t;
            		accept = (((w*(m+y-m*y))<=1) || (w<=(pow(y,m-1))));
				}
			}
		}	
		else{
  			// [Matlab:] Use Best's rejection algorithm XG, p. 410
  			double b = m-1;
  			double c = 3*m-0.75;
			int accept=0;
			while (accept==0){
          		double u = unif_rnd(0,1); 
				double v = unif_rnd(0,1);
          		double w = u*(1-u);  
				double y = sqrt(c/w)*(u-0.5);
          		x = b + y;
          		if (x >= 0){
             		double z = 64*pow(w,3)*v*v;
             		accept = (z<=(1-2*y*y/x)) || (log(z)<=(2*(b*log(x/b)-y)));
          		}
   			}
		}
		double rnd = x;
		return rnd;
	}
	// -------------------
	// Normal distribution.
	// -------------------
	static double normal_pdf(double x){
		double y = exp(-x*x*0.5)/2.506628275;
		return y;
	}
	static double normal_rnd(){
		// Using the inverse of cumulative normal. 
		/*
		double c_0 = 2.515517, c_1 = 0.802853, c_2 = 0.010328;
 		double d_1 = 1.432788, d_2 = 0.189269, d_3 = 0.001308;
		double u = unif_rnd(0, 1);
		bool b = (u > 0.5);
		if (b) u = 1 - u; 
		double t = sqrt(-2*log(u)*0.69314718);
		double y = t - (c_0 + t*(c_1 + c_2*t))/(1 + t*(d_1 + t*(d_2 + d_3*t)));
		if (b) y = -y;
		return y;
		*/
		// The Box-Muller method. 
		/*
		double u1 = unif_rnd(0, 1);
		double u2 = 6.283185307*unif_rnd(0, 1);
		double g = sqrt(-2*LN(u1));
		double x1 = g*cos(u2);
		double x2 = g*sin(u2);
		return x2;
		*/
		
		// The rejection (polar) method. 
		if (is_save_normal){
			is_save_normal = false;
			return save_normal;
		}
		is_save_normal = true;
		double r = 2, u1 = 0, u2 = 0;
		while (r > 1){
			u1 = unif_rnd(-1, 1); u2 = unif_rnd(-1, 1);	
			r = u1*u1 + u2*u2;
		}
		double g = sqrt(-2*LN(r)/r);
		double x = g*u1; save_normal = g*u2;
		return x;
		
	}
	static double normal_cdf(double x){
		if (x < 0) return 1 - normal_cdf(-x);
		double a_1 = 0.4361836, a_2 = -0.1201676, a_3 = 0.9372980;
		double t = 1/(1 + 0.33267*x);
		double y = 1 - normal_pdf(x) * t*(a_1 + t*(a_2 + t*a_3));
		return y; 
	}
	static double normal_pdf(double m, double v, double x){
		double d = sqrt(v); double y = normal_pdf((x-m)/d)/d;
		return y;
	}
	static double normal_rnd(double m, double v){
		double x = normal_rnd();
		double rnd = m + sqrt(v)*x;
		return rnd;
	}	
	
	// -----------------------
	// Test and print methods.
	// -----------------------
	static void test(){
		int n = 64*4000;
		fprintf(stderr, 
			"random > test: a sample of size %d from the N(0,1)\n", n); 
		double x[64*4096];
		clock_t t1 = clock();
		for(int i=2; i<n; i++){
			//x[i] = unif_rnd(0, 1);
			//x[i] = beta_rnd(8, 8);
			//x[i] = gamm_rnd(0.5, 1);
			x[i] = normal_rnd();
		}
		clock_t t2 = clock();
		fprintf(stderr, "speed: %6.3f seconds used\n", 
			((double) (t2-t1)) / CLOCKS_PER_SEC ); 
		//for (double x = 0.0; x < 5; x+=0.01){
		//	fprintf(stdout,"%8.5f, %8.5f %8.5f\n", 
		//		x, normal_pdf(x), normal_cdf(x));
		//}
		print_histogram(x, n);
	}
	static void histogram(double a, double b, 
		int *h, int m, double *x, int n){
		
		for (int j = 0; j < m; j++){
			h[j] = 0;
		}
		for (int i = 0; i < n; i++){
			double y = x[i];
			int j = (int) (m * (y - a)/(b - a));	
			h[j] ++;
		}
	}
	static void print_histogram(double *x, int n){
		int imin = 0, imax = 0;
		for (int i = 1; i < n; i++){
			if (x[i] < x[imin]) imin = i;
			if (x[i] > x[imax]) imax = i;
		}
		int h[32]; int m = 32;
		histogram(x[imin], x[imax], h, m-1, x, n);
		double d = (x[imax] - x[imin])/m;
		double t = n / 512.0;
		fprintf(stderr, "min: %4.2f  max: %4.2f \n", x[imin], x[imax]);
		for (int j = 0; j < m; j++){
			fprintf(stderr, "%5.2f: (%6d) ", x[imin] + j * d, h[j]);
			for (int k = 0; k < h[j]/t; k++){
				fprintf(stderr, "*");
			}
			fprintf(stderr, "\n");
		}  
	}
	
};
#endif
