// This is a program for detecting haplotype blocks from 
// a given set of haplotypes. 
//
// Uses natural logarithms (nats) everywhere.

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <time.h>

#include "Random.h"

// Style. 
//
#define PRINT_NO	1
#define PRINT_S		2
#define PRINT_M		3
#define PRINT_L		4
#define PRINT_STYLE	PRINT_L

#define HGMODE 1 // 1=read alleles as they are, 2=construct genotypes.

// Constants. 
//
#define MAX_HAPLOTYPES 1028
#define MAX_MARKERS 128
#define MAX_VARIANTS 32
#define MAX_SEGMENT_LENGTH 128

// Global variables. 
//
static int allele[MAX_HAPLOTYPES][MAX_MARKERS];
static int haplotypes;
static int markers;
static double p_error = 0.05;

static FILE *fi, *fo;

#define log2(a) log(a)*1.4427
#define MARK 9e99
// Substitutes la += log(1+exp(lb-la)) <=> a = a + b.
#define LOGADD(la,lb) if (la>8e99 || lb-la>100) la=lb; else la+=log(1+exp(lb-la));

void read_data(FILE *f){
	markers = 0;
	int i = 0, j = 0;
	char str[16], c = fgetc(f); 
	int k = 0; 
	while (c != EOF){
		//fprintf(stderr, "%c.", c);
		if (isspace(c)){ 
			//fprintf(stderr, "#.");
			if (k > 0) {str[k] = '\0'; allele[i][j++] = atoi(str); k = 0;}
		}
		else{ 
			//fprintf(stderr, "%c.", c);
			str[k++] = c;
		}
		if (c == '\n'){
			//fprintf(stderr, "[%d:%d]", allele[i][0],allele[i][j-1]);
			if (j > markers) markers = j; i++; j = 0;
		}
		c = fgetc(f);
	}
	haplotypes = i; 
	fprintf(stderr, "data read: %d markers,  %d haplotypes\n", 
		markers, haplotypes); 
}

class Block {
public:
	Block(){
		n = haplotypes; 
		Random::init();
	}
	~Block(){}

	void initialize(int i, int j){
		m = j - i + 1;
		//int variants[m];
		double freq[m][MAX_VARIANTS];
		for (int r = 0; r < m; r++){
			if (HGMODE == 1)
			for (int h = 0; h < n; h++){
				int a = allele[h][i+r];
				data[h][r] = a;
			}
			else { 
				for (int h = 0; h+1 < n; h +=2){
					int a = allele[h][i+r], b = allele[h+1][i+r];
					if (a*b == 0) data[h/2][r] = 0;
					else if (a != b) data[h/2][r] = 3;
					else data[h/2][r] = a;	
				}
				n = n / 2;
			}
		}
		for (int r = 0; r < m; r ++){
			variants[r] = 0;
			int tmp[MAX_VARIANTS];
			// Initialize.
			for (int a = 0; a < MAX_VARIANTS; a ++){
				tmp[a] = 0; freq[r][a] = 0;
			}
			// Count.
			for (int h = 0; h < n; h ++){
				int a = data[h][r];
				tmp[a] ++; freq[r][a] ++;
			}
			// Count.
			double s = 0;
			for (int a = 1; a < MAX_VARIANTS; a ++){
				if (tmp[a] > 0) variants[r] ++;
				s += freq[r][a];
			}
			lrvar[r] = log(variants[r] - 1);
			freq[r][0] = 0; // Count only observed alleles.
			// Normalize.
			for (int a = 1; a < MAX_VARIANTS; a ++){
				freq[r][a] /= s;
				if (freq[r][a] > 0) lfreq[r][a] = log(freq[r][a]);
				else lfreq[r][a] = 999;
			}
		}
		
		/*
		if (i == 1 && j == 63){
			fprintf(stderr, "\n nr of variants at the loci: ");
			for (int r = 0; r < m; r ++){
			fprintf(stderr, " %d", variants[r]);
			}
		}
		*/
	}
	// This is a novel sequential coding model based on a kernel technique.
	// Time requirement O(n^2 m). (MK, Mar 3, 2003) 
	// Note that the returned score should give the 2-based logarithm of 
	// the marginal likelihood P(D[i,j] | isblock[i,j]).
	// Note that this procedure does not provide any estimates
	// for the number of clusters or ancestral haplotypes. All that is
	// implicit and totally hidden.
	void find_score(int a, int b){
		m = b - a + 1;
		initialize(a, b);
		// Code haplotypes one by one. 
		// The probability distribution is a mixture of 
		// (i + 1) distributions.
		int tmp[n];
		for (int i = 0; i < n; i ++) tmp[i] = i;
		Random::rperm(tmp, n);
		
		sc = 0;
		for (int i = 0; i < n; i ++){
			double ls = MARK;
			// Probability of the null model.
			LOGADD(ls, log_probability(tmp[i], -1));
			for (int h = 0; h < i; h ++){
				LOGADD(ls, log_probability(tmp[i], tmp[h]));
				if (ls > MARK) {
					fprintf(stderr, "\nERROR:\n");
					fprintf(stderr, " ls = %e \n", ls);
					fprintf(stderr, " lp(%d,%d) = %e (m=%d) \n", 
						i, h, log_probability(i, h), m);
					exit(1);
				}
			}
			// This many bits are needed to code i-th haplotype given 
			// all the prescending haplotypes.
			score[tmp[i]] = -ls + log(i + 1);
			sc += score[tmp[i]];
			
		}
		asc = sc;
	}
	// Note that these probabilities could be precomputed, 
	// yielding a speed-up for the overall algorithm. 
	// Other speed-ups are also possible. 
	// This is a slower prototype version.
	// Logarithmic version.
	double log_probability(int i, int h){
		double lp_error = log(p_error), lpnoerror = log(1 - p_error);
		double lq = 0;
		for (int j = 0; j < m; j ++){
			if (data[i][j] == 0){}
			else if (h == -1 || data[h][j] == 0)
				lq += lfreq[j][data[i][j]];
			else if (data[i][j] == data[h][j])
				lq += lpnoerror;
			else {
				lq += lp_error - lrvar[j];
				if (variants[j] == 1){
					fprintf(stderr, "*****"); exit(1);
				}
			}
		}
		return lq;
	}
	int get_data(int h, int r){return data[h][r];}
	double get_score(){return sc;}
	double get_avgscore(){return asc;}
	double get_score(int i){ return score[i];}
private:
	double sc; // Score.
	double asc;	// Ave sc over diff nr of clusters.
	int n; // Number of rows.
	int m; // Number of columns.
	int data[MAX_HAPLOTYPES][MAX_MARKERS];
	double lfreq[MAX_MARKERS][MAX_VARIANTS];
	double lrvar[MAX_MARKERS];
	int variants[MAX_MARKERS];
	double score[MAX_HAPLOTYPES];
};

// Computes a probability table prob[i][h];
void initialize(){

}

void scoreit() {
	
	int M = markers;
	double sc, asc;
	Block block;

	for (int i = 0; i < M; i++) {
    	fprintf(stderr, "Block [%3d, j], j =",i);
		for (int j = i; j < M; j++) {
			if ((j-i+1) < MAX_SEGMENT_LENGTH) {
				block.find_score(i, j);
				sc = block.get_score();
				asc = block.get_avgscore();
			} 
			else {
				asc = 9e6; sc = 9e6;
			}
			fprintf(stderr, " %d",j);		
  			fprintf(fo, " %3d %3d  ", i, j);
			fprintf(fo, " %d %d\n", (int)(1.0 + sc), (int)(1.0 + asc)); 
		}
		fprintf(stderr, ".\n");
	}
}

void refineit(){
	FILE *f = fopen("blocks.seg","r");
	Block block;
	bool ok = true;
	int aa[MAX_MARKERS], bb[MAX_MARKERS];
	double score[haplotypes][MAX_MARKERS];
	int s = 0;
	while (ok){
		int a, b;
		int r = fscanf(f, "%d %d", &a, &b);
		ok = (r == 2);
		if (!ok) break;
		if (ok){
			aa[s] = a; bb[s] = b;
			//fprintf(stderr, "Block [%3d, %3d]\n", a, b);
			block.find_score(a, b);
			for (int i = 0; i < haplotypes; i ++){
				score[i][s] = block.get_score(i);
			}
			s ++;
		}
	}
	fclose(f);
	fprintf(stderr, " %d blocks detected.\n", s);
	fprintf(fo, "#  sum ");
	for (int j = 0; j < s; j ++){
		fprintf(fo, " [%3d,%3d]", aa[j], bb[j]);
	}
	fprintf(fo, "\n"); 
	for (int i = 0; i < haplotypes; i ++){
		double sum = 0;
		for (int j = 0; j < s; j ++){
			sum += score[i][j];
		}	
		fprintf(fo, "%6.3f", sum);
		for (int j = 0; j < s; j ++){
			fprintf(fo, "   %7.4f", score[i][j]);
		}
		fprintf(fo, "\n");
	}
}

int main(int argc, char **argv) {
	
	char finame[64], foname[64];
	int i = 0;
	while (argv[2][i] != '\0'){
		finame[i] = foname[i] = argv[2][i];
		i ++;
	}
	finame[i] = foname[i] = '.';
	finame[i+1] = '0'; finame[i+2] = '1'; finame[i+3] = '2'; finame[i+4] = '\0';
	foname[i+1] = 's'; foname[i+2] = 'c'; foname[i+3] = 'o'; foname[i+4] = '\0';
	
	fprintf(stderr, "score > I/O: %s -> %s\n", finame, foname); 
	
	fi = fopen(finame, "r");
	read_data(fi);
	fclose(fi);
	
	fo = fopen(foname, "w");
  	initialize();
	p_error = atof(argv[1]);
	scoreit();	
	fclose(fo);
	
	return 1;
}
